""" File for mathematical structures like a memory polynomial"""

import numpy as np


class MemoryPolynomial:

    def __init__(self, order: int = 5, memory_depth: int = 4, memory_stride: int = 1):
        """Create an instance of a parallel Hammerstein, memory polynomial"""

        self.check_for_errors(order, memory_depth, memory_stride)

        # Save to object
        self.order = order
        self.memory_depth = memory_depth
        self.memory_stride = memory_stride
        self.coeffs = np.zeros((self.n_rows, self.memory_depth))

    def transmit(self, x):
        """Transmit a signal through the Memory Polynomial object"""

        X = self.setup_basis_matrix(x)
        coeffs = self.coeffs.flatten().copy()
        return np.dot(X, coeffs)

    def perform_least_squares(self, x, y):
        """Perform a least squares fit
        Todo:
            - Add support for regularized LS.
        """
        X = self.setup_basis_matrix(x)
        # lamb = 0.0001
        # coeffs, residuals, rank, s = np.linalg.lstsq(X.T.dot(X) + lamb * np.identity(self.n_coeffs),
        # X.T.dot(y), rcond=None)
        coeffs = np.linalg.lstsq(X, y, rcond=None)[0]
        return coeffs

    def setup_basis_matrix(self, x):
        """Setup a matrix of the signal and delayed replicas for multiplication by the coeffs"""
        X = np.zeros((x.size, self.n_coeffs), dtype=np.complex64)
        column_index = 0
        for order in range(1, self.order + 1, 2):
            branch = np.multiply(x, np.power(np.abs(x), (order - 1)))
            for tap in range(0, self.memory_depth):
                delay = tap * self.memory_stride
                X[:, column_index] = np.resize(np.insert(branch, 0, np.zeros(delay)), branch.size)
                column_index += 1
        return X

    @staticmethod
    def check_for_errors(order, memory_depth, memory_stride):
        """Check for errors. Must be odd order with positive memory"""
        if order % 2 == 0 or order <= 0:
            raise Exception("PA Order must be positive and odd")

        if memory_depth <= 0:
            raise Exception("Memorydepth must be positive int")

        if memory_stride <= 0:
            raise Exception("Memory Stride must be a positive int")

    @property
    def n_coeffs(self):
        """"Total number of coefficients including the polynomial order and memory depth"""
        return self.memory_depth * self.n_rows

    @property
    def n_rows(self):
        """Total number of rows in the coeff matrix"""
        return np.floor_divide(self.order + 1, 2)



"""DSP Module for basic DSP functions"""
import numpy as np


def frequency_shift(signal, shift_amount, sampling_rate):
    """Performs a shift in the freqeuency shift by multiplying by a complex sinusoid

    Args:
        signal: The signal to be shifted as a nparray
        shift_amount: Amount to shift by in Hz
        sampling_rate: The original sampling rate of the signal in Hz

    Returns:
        Returns a nparray with the signal shifted by the shift amount
    """
    return signal * np.exp(2*np.pi*1j*np.arange(signal.size)*shift_amount/sampling_rate)



# -*- coding: utf-8 -*-

"""Top-level package for phypy."""

__author__ = """Chance Tarver"""
__email__ = 'tarver.chance@gmail.com'
__version__ = '0.2.7'

from . import analog
from . import corrections
from . import modulators
from . import dsp

__all__ = ['analog', 'corrections', 'modulators', 'dsp', 'structures']



# -*- coding: utf-8 -*-

"""Module for performing 'analog' related PHY tasks such as the power amplifier"""
import numpy as np
try:
    from .structures import MemoryPolynomial
except:
    from structures import MemoryPolynomial


class PowerAmp(MemoryPolynomial):
    """Power amplifier class that implements a baseband, memory-polynomial based PA """

    def __init__(self, order: int = 5, memory_depth: int = 4, memory_stride: int = 1,
                 noise_variance: float = 0.05, add_lo_leakage: bool = True,
                 add_iq_imbalance: bool = True, seed: int = 1):
        """Creates an instance of a parallel Hammerstein PA model extracted from a WARP PA board"""

        super().__init__(order, memory_depth, memory_stride)

        # Seed the random number generator for reproducibility
        np.random.seed(seed)

        if noise_variance < 0:
            raise Exception("The noisevarriance must be >=0")
        else:
            self.noise_variance = noise_variance

        if add_lo_leakage:
            self.lo_leakage = 0.01*np.random.randn() + 0.01j*np.random.randn()
        else:
            self.lo_leakage = 0

        if add_iq_imbalance:
            gm = 1.07
            pm = 5*np.pi/180
            k1 = 0.5*(1 + gm * np.exp(1j*pm))
            k2 = 0.5*(1 - gm * np.exp(1j*pm))
            sc_iq = 1/np.sqrt(np.abs(k1)**2 + np.abs(k2)**2)
            self.k1 = sc_iq*k1
            self.k2 = sc_iq*k2
        else:
            self.k1 = 1
            self.k2 = 0
        default_poly_coeffs = np.array([[0.9295 - 0.0001j, 0.2939 + 0.0005j, -0.1270 + 0.0034j, 0.0741 - 0.0018j],    # 1st order coeffs
                                        [0.1419 - 0.0008j, -0.0735 + 0.0833j, -0.0535 + 0.0004j, 0.0908 - 0.0473j],   # 3rd order
                                        [0.0084 - 0.0569j, -0.4610 + 0.0274j, -0.3011 - 0.1403j, -0.0623 - 0.0269j],  # 5th order
                                        [0.1774 + 0.0265j, 0.0848 + 0.0613j, -0.0362 - 0.0307j, 0.0415 + 0.0429j]],   # 7th order
                                       np.complex64)

        self.coeffs = default_poly_coeffs[:self.n_rows, :self.memory_depth]
        self.nmse_of_fit = None  # In case we fit the PA to some model

    def transmit(self, x):
        x = self.k1*x + self.k2*np.conj(x)
        return super().transmit(x) + self.noise_variance*np.random.rand(x.size)

    def make_new_model(self, pa_input, pa_output):
        """Learn new coefficients based on pa_inputs and pa_outputs"""
        self.coeffs = self.perform_least_squares(pa_input, pa_output).reshape(self.coeffs.shape)
        model_pa_output = self.transmit(pa_input)
        self.nmse_of_fit = self.calculate_nmse(pa_output, model_pa_output)

    @staticmethod
    def calculate_nmse(desired, actual):
        """Calculate the normalized mean squared error
        Todo:
            - Check this. I think I need to divide by number of samples
        """
        return np.linalg.norm(desired-actual)**2 / np.linalg.norm(desired)**2


if __name__ == "__main__":
    import matplotlib.pyplot as plt

    pa = PowerAmp(order=7, noise_variance=0, add_iq_imbalance=False, add_lo_leakage=False, memory_stride=2)
    pa.coeffs = np.zeros(pa.coeffs.shape)
    pa.coeffs[(0, 0)] = 1
    x = np.exp(1j * (2 * np.pi * 1e6 * np.arange(100)/10e6))
    y = pa.transmit(x)
    pa.make_new_model(x, y)

    plt.plot(x.real)
    plt.plot(y.real)
    plt.show()



"""Module for performing corrections on impairments related  to the PHY such as DPD"""
import numpy as np
try:
    from .structures import MemoryPolynomial
except:
    from structures import MemoryPolynomial


class ILA_DPD(MemoryPolynomial):
    """Implements a DPD object that uses an indirect learning architecture (ILA)

    Implements a digital predistorter (DPD) that uses an indirect learning architecture (ILA)
    and a parallel hammerstein, memory polynomial structure that acts as an inverse of the PA model.

    """
    def __init__(self, order: int = 5, memory_depth: int = 1, memory_stride: int = 5, n_iterations: int = 2):
        self.n_iterations = n_iterations

        super().__init__(order, memory_depth, memory_stride)

        # Make the 1st coeff 1 to have a completely linear DPD with 0 effect
        self.coeffs = np.zeros(shape=(self.n_rows, self.memory_depth))
        self.coeffs[0, 0] = 1

    def perform_learning(self, pa, signal):
        """Learn a new DPD model for a given pa"""
        for _ in range(self.n_iterations):
            # Forward through the predistorter
            pa_input = self.transmit(signal)

            # Transmit the predistorted signal through the actual PA
            pa_output = pa.transmit(pa_input)

            # Remove any PA Gain
            pa_output = pa_output*np.linalg.norm(pa_input)/np.linalg.norm(pa_output)

            # Learn on the postdistorter
            self.coeffs = self.perform_least_squares(pa_output, pa_input).reshape(self.coeffs.shape)


if __name__ == "__main__":
    import matplotlib.pyplot as plt
    import modulators
    import analog

    dpd = ILA_DPD()

    ofdm = modulators.OFDM(n_subcarriers=600)
    x = ofdm.use(n_symbols=4)

    x = x*15

    pa = analog.PowerAmp(add_iq_imbalance=False, add_lo_leakage=False, noise_variance=0)
    pa.coeffs = np.zeros(shape=pa.coeffs.shape)
    pa.coeffs[0, 0] = 2
    pa.coeffs[1, 0] = 0.1
    pa.coeffs[2, 0] = -0.05
    y = pa.transmit(x)
    dpd.perform_learning(pa, x)
    pa_input_with_dpd = dpd.transmit(x)
    pa_output_with_dpd = pa.transmit(pa_input_with_dpd)

    # Analyze results.
    plt.psd(x, 1024, 10e6, label='original')
    plt.psd(y, 1024, 10e6, label='no dpd')
    plt.psd(pa_output_with_dpd, 1024, 10e6, label='with dpd')
    plt.gca().set_ylim(bottom=-120)
    plt.legend()
    plt.show()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `phypy` package."""

import pytest
import numpy as np
from phypy import analog
from phypy import modulators as mods
from phypy import dsp

def test_pa_setup():
    pa = analog.PowerAmp(order=5, memory_depth=2, noise_variance=0.01, add_lo_leakage=False, add_iq_imbalance=False)
    assert pa.order == 5
    assert pa.memory_depth == 2
    assert pa.noise_variance == 0.01
    assert pa.lo_leakage == 0
    assert pa.k1 == 1
    assert pa.k2 == 0


def test_pa_transmission_with_unit_coeff():
    """Test that the pa output is equal to the pa input if there is only a 1 in the 1st order term"""
    pa = analog.PowerAmp(order=7, noise_variance=0, add_iq_imbalance=False, add_lo_leakage=False)
    pa.coeffs = np.zeros(pa.coeffs.shape)
    pa.coeffs[(0, 0)] = 1
    x = np.exp(1j * (2 * np.pi * 1e6 * np.arange(100)/10e6))
    assert sum(np.abs(pa.transmit(x) - x)) <= 1e-5


def test_ofdm_setup_lte_20mhz():
    """Test the default setup of the OFDM class"""
    ofdm = mods.OFDM()
    assert ofdm.n_subcarriers == 1200
    assert ofdm.subcarrier_spacing == 15000  # 15 kHz
    assert ofdm.cp_length == 144  # Normal cyclic prefix
    assert ofdm.fft_size == 2048
    assert ofdm.sampling_rate == 30.72e6  # 30.72 MHz
    assert (ofdm.symbol_alphabet == np.array([-1+1j, -1-1j, 1+1j, 1-1j])).all()


def test_ofdm_setup_lte_5mhz():
    """Test the default setup of the OFDM class"""
    ofdm = mods.OFDM(n_subcarriers=300)
    assert ofdm.n_subcarriers == 300
    assert ofdm.subcarrier_spacing == 15000  # 15 kHz
    assert ofdm.cp_length == 144  # Normal cyclic prefix
    assert ofdm.fft_size == 512
    assert ofdm.sampling_rate == 7.68e6  # MHz
    assert (ofdm.symbol_alphabet == np.array([-1+1j, -1-1j, 1+1j, 1-1j])).all()


def test_freq_shift():
    """ Test the frequency shift. A 1 MHz signal shifted by 1 MHz should be a 2 MHz signal"""
    sampling_rate = 20e6  # 20 MHz
    sin_freq = 1e6  # 1 MHz sinusoid
    seconds = 0.00001  # 0.01 ms of data

    sampling_period = 1/sampling_rate
    t_array = np.arange(start=0, stop=seconds, step=sampling_period)
    samples = np.exp(2*np.pi*1j*sin_freq*t_array)
    y = dsp.frequency_shift(samples, shift_amount=1e6, sampling_rate=sampling_rate)
    two_mhz_sin = np.exp(2*np.pi*1j*(2*sin_freq)*t_array)
    error = np.square(np.abs(two_mhz_sin - y)).mean()
    assert error < 1e-20



"""Modulator Module that implements various wireless modulators

These modulators are meant to turn arbitrary bit patterns to analog waveforms for wireless
transmission.
"""

import numpy as np


class OFDM:
    """Class that creates OFDM signals.

    This class will set up an OFDM modulator to create random OFDM signals.

    Attributes:
        n_subcarriers: Number of subcarriers per OFDM symbol
        subcarrier_spacing: Spacing between subcarriers in Hz
        cp_length : Number of samples in the cyclic prefix
        fft_size: Size of the IFFT/FFT used.
        sampling_rate: The native sampling rate based on the FFT size and subcarrier spacing
        symbol_alphabet: The constellation points

    Todo:
        - Add an arbitrary bit input
        - Add a demodulator
    """

    def __init__(self, n_subcarriers: int = 1200, subcarrier_spacing: int = 15000,
                 cp_length: int = 144, constellation: str = 'QPSK', seed: int = 0):
        """OFDM Modulator Constructor.

        Construct an OFDM Modulator with custom number of subcarriers, subcarrier spacing,
        cyclic prefix length, and constellation on each subcarrier.

        Args:
            n_subcarriers: Number of subcarriers per OFDM symbol
            subcarrier_spacing: Spacing of the subcarriers in the frequency domain in Hertz
            cp_length: Number of samples in cyclic prefix
            constellation: Type of constellation used on each subcarrier. QPSK, 16QAM or 64QAM
            seed: Seed for the random number generator
        """
        self.n_subcarriers = n_subcarriers
        self.subcarrier_spacing = subcarrier_spacing
        self.cp_length = cp_length

        self.fft_size = np.power(2, np.int(np.ceil(np.log2(n_subcarriers))))
        self.sampling_rate = self.subcarrier_spacing * self.fft_size
        self.symbol_alphabet = self.qam_alphabet(constellation)
        self.seed = seed
        self.fd_symbols = None  # We'll hold the last TX symbols for calculating error later

    def use(self, n_symbols: int = 10):
        """Use the OFDM modulator to generate a random signal.

        Args:
            n_symbols: Number of OFDM symbols to generate

        Returns:
            A time-domain OFDM signal

        TODO:
            - Allow to pass in an arbitrary bit pattern for modulation.
        """
        np.random.seed(self.seed)
        self.fd_symbols = self.symbol_alphabet[
            np.random.randint(self.symbol_alphabet.size, size=(self.n_subcarriers, n_symbols))]
        out = np.zeros((self.fft_size + self.cp_length, n_symbols), dtype='complex64')
        for index, symbol in enumerate(self.fd_symbols.T):
            td_waveform = self.frequency_to_time_domain(symbol)
            out[:, index] = self.add_cyclic_prefix(td_waveform)

        return out.flatten(1)

    def frequency_to_time_domain(self, fd_symbol):
        """Convert the frequency domain symbol to time domain via IFFT

        Args:
            fd_symbol: One frequency domain symbol

        Returns:
            time domain signal
        """
        # TODO: Verify that the RB are mapping to the IFFT input correctly
        ifft_input = np.zeros((self.fft_size), dtype='complex64')
        # Index 0 is DC. Leave blank. The 1st half needs to be in negative frequency
        # so they go in the last IFFT inputs.
        ifft_input[1: np.int(self.n_subcarriers / 2) + 1] = \
            fd_symbol[np.int(self.n_subcarriers / 2):]
        ifft_input[-np.int(self.n_subcarriers / 2):] = \
            fd_symbol[:np.int(self.n_subcarriers / 2)]
        return np.fft.ifft(ifft_input)

    def time_to_frequency_domain(self, td_symbol):
        full_fft_output = np.fft.fft(td_symbol, axis=0)
        fd_symbols = np.zeros(shape=self.fd_symbols.shape, dtype='complex64')
        fd_symbols[np.int(self.n_subcarriers / 2):, :] = full_fft_output[1:np.int(self.n_subcarriers/2) + 1, :]
        fd_symbols[:np.int(self.n_subcarriers / 2), :] = full_fft_output[-np.int(self.n_subcarriers / 2):, :]
        return fd_symbols

    def add_cyclic_prefix(self, td_waveform):
        """Adds cyclic prefix

        Adds by taking the last few samples and appending it to the beginning of the signal

        Args:
            td_waveform: IFFT output signal.

        Returns:
            time domain signal with a cyclic prefix
        """

        # TODO: verify my indexing
        out = np.zeros(td_waveform.size + self.cp_length, dtype='complex64')
        out[self.cp_length:] = td_waveform
        out[:self.cp_length] = td_waveform[-self.cp_length:]
        return out

    def remove_cyclic_prefix(self, td_grid):
        w_out_cp = td_grid[-self.fft_size:, :]
        return w_out_cp

    @staticmethod
    def qam_alphabet(constellation):
        """Returns constellation points for QPSK, 16QAM, or 64 QAM

        Args:
            constellation: String saying the desired constellation

        Returns:
            symbol alphabet on the complex plane
        """
        constellation_dict = {
            "QPSK": 4,
            "16QAM": 16,
            "64QAM": 64
        }
        n_points = constellation_dict[constellation]
        x = np.int(np.sqrt(n_points)) - 1

        alpha_n_points = np.arange(-x, x + 1, 2, dtype=int)
        A = np.kron(np.ones((x + 1, 1)), alpha_n_points)
        B = np.flipud(A.transpose())
        const_qam = A + 1j * B
        alphabet = const_qam.flatten(1)
        return alphabet


    def demodulate(self, time_domain_rx_signal):
        """Demodulate a time domain signal back into the FD symbols"""

        # Reorganize to grid
        _, n_symbols = self.fd_symbols.shape
        td_grid = np.reshape(time_domain_rx_signal, (self.fft_size+self.cp_length, n_symbols), order='F')
        td_grid = self.remove_cyclic_prefix(td_grid)
        fd_symbols = self.time_to_frequency_domain(td_grid)
        evm = self.calculate_evm(fd_symbols)
        return fd_symbols, evm

    def calculate_evm(self, fd_rx_signal):
        # Get error vectors
        e = fd_rx_signal - self.fd_symbols
        evm = 100 * np.linalg.norm(e) / np.linalg.norm(self.fd_symbols)
        return evm

if __name__ == "__main__":
    ofdm = OFDM()
    x = ofdm.use()
    y, evm_percent = ofdm.demodulate(x)
    1 + 1
